#! /usr/bin/env Rscript

## Functions
geom.mean <- function(x, na.rm = T, zero.propagate = F) {
    exp(mean(log(x + 1), na.rm = na.rm)) - 1
}

geom.sd <- function(x, na.rm = T) {
    exp(sd(log(x + 1), na.rm = na.rm))
}

## --------------- Start of Script ---------------
args <- commandArgs(T)

if (length(args) < 1) {
    stop('Usage: Rscript crunch_exp1.r <data-file>')
}

require('ggplot2')
require('methods')

## --------------- Script Parameters ---------------
base.stats <- c('WeightedCost', 'Depth', 'Gates', 'AllocTime', 'BMTPartitions')
type.map <- c(WeightedCost = 'cost', Depth = 'depth', Gates = 'gates', AllocTime = 'time', BMTPartitions = 'ptt')
base.params <- list(c(32, 2560))

## --------------- Preprocessing ---------------
data <- read.table(args[1], header = T, sep = ',')
data$Benchmark <- gsub('.qasm', '', as.character(data$Benchmark))

data <- aggregate(cbind(AllocTime, WeightedCost, Depth, Gates, BMTPartitions)
                          ~ Allocator + Benchmark + 
                              MaxChildren + MaxPartial,
                  data,
                  function(v) c(Mean = mean(v), Sd = sd(v), Min = min(v)))

for (stat in base.stats) {
    data[[stat]] <- data.frame(data[[stat]])
}

## --------------- Processing Each Stat ---------------
max.children <- sort(unique(data$MaxChildren))
max.partial <- sort(unique(data$MaxPartial))

header <- " & "
for (mp in max.partial) {
    header <- paste(header, mp, sep = " & ")
}


for (base in base.params) {
    base.mc <- base[1]
    base.mp <- base[2]

    ## --------------- Header of the TeX file ---------------
    latex.file <- paste('Figure17-baseline.', base.mc, '.', base.mp, '.tex', sep = '')

    cat('\\documentclass[10pt]{article}\n', file = latex.file)
    cat('\\usepackage{pdflscape}\n', file = latex.file, append = T)
    cat('\\usepackage{amsmath}\n', file = latex.file, append = T)
    cat('\\usepackage{multirow}\n', file = latex.file, append = T)
    cat('\\begin{document}\n', file = latex.file, append = T)
    cat('\\begin{landscape}\n', file = latex.file, append = T)

    cat("\\begin{figure}\n", file = latex.file, append = T)
    cat("\\centering\n", file = latex.file, append = T)
    cat("\\begin{tabular}{| c || c |  c | c | c | c | c | c | c | c | c |}\n",
        file = latex.file, append = T)
    cat("\\hline\n", file = latex.file, append = T)
    cat(paste(header, "\\\\\n"), file = latex.file, append = T)
    cat("\\hline\n", file = latex.file, append = T)
    cat("\\hline\n", file = latex.file, append = T)
    ## ------------------------------------------------------

    base.data <- data[data$MaxChildren == base.mc & data$MaxPartial == base.mp, ]
    for (mc in max.children) {
        cat(paste("\\multirow{", length(base.stats), "}{*}{", mc, "}\n"), file = latex.file, append = T)

        for (stat in base.stats) {
            row <- c(type.map[stat])

            for (mp in max.partial) {
                comp.data <- data[data$MaxChildren == mc & data$MaxPartial == mp, ]
                bench.order <- match(base.data$Benchmark, comp.data$Benchmark)
                data.ratio <- comp.data[bench.order, stat]$Mean / base.data[, stat]$Mean
                data.geomean <- sprintf("%.2f", geom.mean(data.ratio))
                data.geosd <- sprintf("%.2f", geom.sd(data.ratio))
                row <- c(row, paste(data.geomean, ' (', data.geosd, ')', sep = ''))
            }

            cat(paste(" &", paste(row, collapse = " & ")), '\\\\\n', file = latex.file, append = T)
        }

        cat("\\hline\n", file = latex.file, append = T)
    }

    ## --------------- Footer of the TeX file ---------------
    cat("\\end{tabular}\n", file = latex.file, append = T)
    cat(paste("\\caption{Exploration of the space of parameters. ", 
              "Rows group results by the maximum number of children ",
              "($M_c$ ), and columns group results by maximum number of ",
              "partial solutions ($M_p$ ). Row ptt denotes the number ",
              "of partitions in which we split the quantum circuit, ",
              "given a certain set of parameters.}\n", sep = ""),
        file = latex.file, append = T)
    cat("\\end{figure}\n", file = latex.file, append = T)

    cat('\\end{landscape}\n', file = latex.file, append = T)
    cat('\\end{document}\n', file = latex.file, append = T)
    ## ------------------------------------------------------
}

warnings()
