#! /usr/bin/env Rscript

## Functions
geom.mean <- function(x, na.rm = T, zero.propagate = F) {
    if (any(x < 0, na.rm = T)) {
        return(NaN)
    }

    if (zero.propagate) {
        if (any(x == 0, na.rm = T)) {
            return(0)
        }

        exp(mean(log(x), na.rm = na.rm))
    } else {
        exp(sum(log(x[x > 0]), na.rm = na.rm) / length(x))
    }
}

geom.sd <- function(x, na.rm = T) {
    exp(sd(log(x), na.rm = na.rm))
}

na_is_zero <- function(x) {
    x[is.na(x)] = 0
    x[!is.na(x)] = round(x, 4)
    return(x)
}

nan_is_dash <- function(x) {
    if (is.nan(x)) {
        return("-")
    } else {
        return(round(x, 4))
    }
}

escape.latex <- function(x) {
    return(gsub("_", "\\\\_", x))
}

log10.label <- function(x) {
    lg <- floor(log10(x))
    pw <- 10 ** lg
    rem <- na_is_zero(x - pw)

    x.filter <- !is.na(x) & x != 0
    x.zero.filter <- !is.na(x) & x == 0

    str <- rep('', length(rem))
    str[rem != 0] <- sprintf('%.4f * ', rem[rem != 0])
    str[x.filter] <- paste(str[x.filter], '10^', lg[x.filter], sep = '')
    str[x.zero.filter] <- '0'
    str[is.na(x)] <- 'NA'

    return(parse(text = str))
}

against_stats <- function(against, stat, data.against, data.rest, print.latex, print.latex.file) {
    header <- paste("Allocator", "G. Mean", "G. Std. D.", "Better Count", "Better G. Mean", sep = " & ")
    header <- paste(header, "Worse Count", "Worse G. Mean", sep = " & ")

    print(paste("----------= Against:", against, "=----------"))
    print(paste("> Stat:", stat))
    if (print.latex) {
        cat("\\begin{table}\n", file = print.latex.file, append = T)
        cat("\\centering\n", file = print.latex.file, append = T)
        cat("\\begin{tabular}{| c || c | c | c | c | c | c | c |}\n", file = print.latex.file, append = T)
        cat("\\hline\n", file = print.latex.file, append = T)
        cat(paste(header, "\\\\\n"), file = print.latex.file, append = T)
        cat("\\hline\n", file = print.latex.file, append = T)
        cat("\\hline\n", file = print.latex.file, append = T)
    }

    data.against.order <- order(data.against$BenchId)

    for (alloc in unique(data.rest$Allocator)) {
        data.alloc <- data.rest[grepl(alloc, data.rest$Allocator), ]
        data.alloc.order <- order(data.alloc$BenchId)

        data.mean.ratio <- data.alloc[, stat]$Mean[data.alloc.order] /
            data.against[, stat]$Mean[data.against.order]

        data.min.ratio <- data.alloc[, stat]$Min[data.alloc.order] /
            data.against[, stat]$Min[data.against.order]

        # data.sd <- data.alloc[, stat]$Sd / data.alloc[, stat]$Mean
        data.sd <- geom.sd(data.mean.ratio)

        all.count <- length(benchmark.level)
        better.count <- na_is_zero(table(data.mean.ratio < 1)['TRUE'])
        worse.count <- na_is_zero(table(data.mean.ratio > 1)['TRUE'])

        if (print.latex) {
            row <- paste(alloc, na_is_zero(geom.mean(data.mean.ratio)), sep = " & ")
            row <- paste(row, na_is_zero(geom.mean(data.sd)), sep = " & ")
            row <- paste(row, paste(better.count, " ($", round(100 * better.count / all.count, 2), "\\%$)", sep = ""), sep = " & ")
            row <- paste(row, nan_is_dash(geom.mean(data.mean.ratio[data.mean.ratio < 1])), sep = " & ")
            row <- paste(row, paste(worse.count, " ($", round(100 * worse.count / all.count, 2), "\\%$)", sep = ""), sep = " & ")
            row <- paste(row, nan_is_dash(geom.mean(data.mean.ratio[data.mean.ratio > 1])), sep = " & ")
            cat(paste(escape.latex(row), "\\\\\n"), file = print.latex.file, append = T)
            cat("\\hline\n", file = print.latex.file, append = T)
        } else {
            min.better.count <- na_is_zero(table(data.min.ratio < 1)[2])
            min.worse.count <- na_is_zero(table(data.min.ratio > 1)[2])

            print(paste(alloc, 'Geometric Mean:', na_is_zero(geom.mean(data.mean.ratio))))
            print(paste(alloc, 'G. Standard Deviation:', na_is_zero(geom.mean(data.sd))))
            print(paste(alloc, 'Better Count:', paste(better.count, " (", round(100 * better.count / all.count, 2), "%)", sep = "")))
            print(paste(alloc, 'Better G. Mean:', nan_is_dash(geom.mean(data.mean.ratio[data.mean.ratio < 1]))))
            print(paste(alloc, 'Worse Count:', paste(worse.count, " (", round(100 * worse.count / all.count, 2), "%)", sep = "")))
            print(paste(alloc, 'Worse G. Mean:', nan_is_dash(geom.mean(data.mean.ratio[data.mean.ratio > 1]))))
            print(paste(alloc, 'Min Arithmetic Mean:', na_is_zero(geom.mean(data.min.ratio))))
            print(paste(alloc, 'Min Better Count:', paste(min.better.count, " (", round(100 * min.better.count / all.count, 2), "%)", sep = "")))
            print(paste(alloc, 'Min Better G. Mean:', nan_is_dash(geom.mean(data.min.ratio[data.min.ratio < 1]))))
            print(paste(alloc, 'Min Worse Count:', paste(min.worse.count, " (", round(100 * min.worse.count / all.count, 2), "%)", sep = "")))
            print(paste(alloc, 'Min Worse G. Mean:', nan_is_dash(geom.mean(data.min.ratio[data.min.ratio > 1]))))
        }
    }

    if (print.latex) {
        cat("\\end{tabular}\n", file = print.latex.file, append = T)
        cat(paste("\\caption{\\textbf{", stat, "} ratios against\\textit{",
                  escape.latex(against), "}allocator.}\n"),
            file = print.latex.file, append = T)
        cat("\\end{table}\n", file = print.latex.file, append = T)
    }
}

## --------------- Start of Script ---------------
args <- commandArgs(T)

if (length(args) < 1) {
    stop('Usage: ./plot.r <data-file>')
}

require('ggplot2')
require('methods')

## --------------- Script Parameters ---------------
refs <- c('bmtS')
alloc.list <- c('jku', 'sbr', 'bmtS', 'bmtF', 'chw', 'wpm', 'ibm')
base.stats <- c('WeightedCost', 'Depth', 'Gates', 'AllocTime')
rename.stat.map <- list(c('AllocTime', 'time'), c('WeightedCost', 'cost'),
                        c('Depth', 'depth'), c('Gates', 'gates'))
shape.values <- c(3:4, 21:24)
font.size <- 9
bench.chosen.step <- 6
line.size <- 0.5
point.fill <- 'white'
sd.times <- 2

## --------------- Preprocessing ---------------
data <- read.table(args[1], header = T, sep = ',')
data$Benchmark <- gsub('.qasm', '', basename(as.character(data$Benchmark)))

# BenchId column
data <- cbind2(data, rep(0, nrow(data)))
colnames(data)[length(colnames(data))] <- 'BenchId'
data$Allocator <- as.character(data$Allocator)

# Aggregating data
if ('Type' %in% colnames(data)) {
    data$Allocator <- paste(data$Allocator, data$Type, sep = ".")
    alloc.list <- paste(alloc.list, data$Type, sep = '.')
}

data <- aggregate(cbind(AllocTime, WeightedCost, Depth, Gates)
                          ~ Allocator + Benchmark + BenchId,
                  data,
                  function(v) c(Mean = mean(v), Sd = sd(v), Min = min(v)))

for (stat in base.stats) {
    data[[stat]] <- data.frame(data[[stat]])
}

## --------------- Processing Each Stat ---------------
for (ref.alloc in refs) {
    data.ref <- data[grepl(ref.alloc, data$Allocator), ]
    data.rest <- data[data$Allocator %in% alloc.list[!grepl(ref.alloc, alloc.list)], ]

    for (stat in base.stats) {
        stat.data.ref <- data.ref
        stat.data.rest <- data.rest

        benchmark.order <- order(stat.data.ref[, stat]$Mean)
        benchmark.level <- unique(stat.data.ref$Benchmark[benchmark.order])

        stat.data.ref$BenchId <- match(stat.data.ref$Benchmark, benchmark.level)
        stat.data.rest$BenchId <- match(stat.data.rest$Benchmark, benchmark.level)

        bench.chosen <- c()
        data.ratio <- cbind2(stat.data.rest, rep(0, nrow(stat.data.rest)))
        data.ratio <- cbind2(data.ratio, rep(0, nrow(stat.data.rest)))
        data.ratio.colnames <- colnames(data.ratio)
        colnames(data.ratio)[length(data.ratio.colnames) - 1] <- 'Ratio'
        colnames(data.ratio)[length(data.ratio.colnames)] <- 'RatioMin'

        rest.list <- unique(data.ratio$Allocator)
        ratio.prof <- t(data.frame(1:length(rest.list), 1:length(rest.list)))
        colnames(ratio.prof) <- rest.list

        for (alloc in rest.list) {
            filter <- data.ratio$Allocator == alloc

            data.ratio[filter, ]$Ratio <- data.ratio[filter, stat]$Mean /
                data.ref[, stat]$Mean
            data.ratio[filter, ]$RatioMin <- data.ratio[filter, stat]$Min /
                data.ref[, stat]$Min

            ratio.mean <- geom.mean(data.ratio[filter, ]$Ratio)
            ratio.sd <- geom.sd(data.ratio[filter, ]$Ratio)
            lower <- ratio.mean
            upper <- ratio.mean

            for (i in 1:sd.times) {
                lower <- lower / ratio.sd
                upper <- upper * ratio.sd
            }

            for (bench in unique(data.ratio$Benchmark)) {
                filter.bench <- filter & data.ratio$Benchmark == bench

                if (data.ratio[filter.bench, ]$Ratio < lower ||
                    data.ratio[filter.bench, ]$Ratio > upper) {
                    bench.chosen <- c(bench.chosen, bench)
                }
            }
        }

        bench.chosen <- unique(bench.chosen)

        y.minval <- min(c(0, data[, stat]$Mean))
        y.breaks <- seq(floor(log10(min(data[, stat]$Mean))), ceiling(log10(max(data[, stat]$Mean))), 1)

        plot <- ggplot(data.ratio, aes(x = BenchId, y = Ratio)) +
            scale_x_continuous(expand = c(0.01, 0.01)) +
            scale_y_continuous(name = paste(stat, 'Ratio'), breaks = seq(0, max(data.ratio$Ratio) + 1),
                               limits = c(0, max(data.ratio$Ratio) + 1)) +
            geom_line(aes(group = Allocator, color = Allocator)) +
            geom_hline(yintercept = 1) +
            geom_point(data = data.ratio[data.ratio$Benchmark %in% bench.chosen, ],
                       aes(shape = Allocator),
                       alpha = 0.8,
                       fill = 'white',
                       size = 4) +
            scale_shape_manual(values = shape.values) +
            xlab('Benchmark Collection') +
            theme(
                axis.title.x = element_text(size = font.size),
                axis.title.y = element_blank(),
                axis.text.x = element_blank(),
                axis.text.y = element_text(size = font.size),
                axis.line = element_line(color = 'black'),
                plot.margin = margin(5, 5, 5, 5),
                legend.text = element_text(size = font.size),
                legend.box.background = element_rect(color = 'black', size = 1),
                legend.margin = margin(2, 5, 2, 5),
                legend.key.size = unit(0.7, 'cm'),
                legend.position = c(0.8, 0.85),
                legend.direction = 'horizontal',
                legend.title = element_blank()
                )

        fig.number = switch(stat, "WeightedCost" = 18, "Depth" = 19, "Gates" = 20, -1)
        if (fig.number != -1) {
            ggsave(paste('Figure', fig.number, '-baseline.', ref.alloc, '.pdf', sep = ''), plot, height = 4, width = 10)
        }

        plot <- ggplot(stat.data.rest, aes_string(x = 'BenchId', y = paste(stat, 'Mean', sep = '$'))) +
            scale_x_continuous(expand = c(0.01, 0.01)) +
            scale_y_log10(breaks = 10 ** y.breaks, labels = log10.label) +
            # geom_line(aes(group = Allocator, color = Allocator)) +
            geom_line(aes(group = Allocator, color = Allocator)) +
            geom_point(data = stat.data.rest[stat.data.rest$Benchmark %in% bench.chosen, ],
                       # aes(shape = Allocator, color = Allocator, fill = Allocator),
                       # aes(shape = Allocator, color = Allocator),
                       aes(shape = Allocator),
                       alpha = 0.8,
                       fill = 'white',
                       size = 4) +
            scale_shape_manual(values = shape.values) +
            geom_ribbon(aes_string(ymin = 'y.minval',
                                   ymax = paste(stat, 'Mean', sep = '$'),
                                   alpha = '0.3'),
                        stat.data.ref,
                        show.legend = F) +
            ylab(stat) + xlab('Benchmark Collection') +
            theme(
                axis.title.x = element_text(size = font.size),
                axis.title.y = element_blank(),
                axis.text.x = element_blank(),
                axis.text.y = element_text(size = font.size),
                axis.line = element_line(color = 'black'),
                plot.margin = margin(5, 5, 5, 5),
                legend.text = element_text(size = font.size),
                legend.box.background = element_rect(color = 'black', size = 1),
                legend.margin = margin(2, 5, 2, 5),
                legend.key.size = unit(0.7, 'cm'),
                legend.position = c(0.2, 0.85),
                legend.direction = 'horizontal',
                legend.title = element_blank()
                )

        fig.number = switch(stat, "AllocTime" = 21, -1)
        if (fig.number != -1) {
            ggsave(paste('Figure', fig.number, '-baseline.', ref.alloc, '.pdf', sep = ''), plot, height = 4, width = 10)
        }
    }

    latex.file <- paste('Figure16-baseline.', ref.alloc, '.tex', sep = '')
    cat('\\documentclass[10pt]{article}\n', file = latex.file)
    cat('\\usepackage{pdflscape}\n', file = latex.file, append = T)
    cat('\\usepackage{amsmath}\n', file = latex.file, append = T)
    cat('\\usepackage{multirow}\n', file = latex.file, append = T)
    cat('\\begin{document}\n', file = latex.file, append = T)
    cat('\\begin{landscape}\n', file = latex.file, append = T)

    cat("\\begin{table}\n", file = latex.file, append = T)
    cat("\\centering\n", file = latex.file, append = T)
    cat("\\begin{tabular}{| c || c | c | c | c | c | c | c |}\n",
        file = latex.file, append = T)
    cat("\\hline\n", file = latex.file, append = T)

    header <- paste("Allocator", "Dim", sep = " & ")
    header <- paste(header, "G. Mean", "G. Std. D.", "Better Count", "Better G. Mean", sep = " & ")
    header <- paste(header, "Worse Count", "Worse G. Mean", sep = " & ")
    cat(paste(header, "\\\\\n"), file = latex.file, append = T)

    cat("\\hline\n", file = latex.file, append = T)
    cat("\\hline\n", file = latex.file, append = T)

    data.ref.order <- order(data.ref$BenchId)

    for (alloc in unique(data.rest$Allocator)) {
        data.alloc <- data.rest[grepl(alloc, data.rest$Allocator), ]
        data.alloc.order <- order(data.alloc$BenchId)

        all.count <- length(benchmark.level)

        for (i in 1:length(base.stats)) {
            stat = base.stats[i]

            data.mean.ratio <- data.alloc[, stat]$Mean[data.alloc.order] /
                data.ref[, stat]$Mean[data.ref.order]

            data.min.ratio <- data.alloc[, stat]$Min[data.alloc.order] /
                data.ref[, stat]$Min[data.ref.order]

            data.sd <- geom.sd(data.mean.ratio)
            better.count <- na_is_zero(table(data.mean.ratio < 1)['TRUE'])
            worse.count <- na_is_zero(table(data.mean.ratio > 1)['TRUE'])

            row <- ""

            for (pair in rename.stat.map) {
                if (stat == pair[1]) {
                    row <- paste("&", pair[2])
                    break
                }
            }

            if (i == length(base.stats)) {
                row <- paste("\\multirow{-5}{*}{", alloc, "}\n", row, sep = "")
            }

            row <- paste(row, na_is_zero(geom.mean(data.mean.ratio)), sep = " & ")
            row <- paste(row, na_is_zero(geom.mean(data.sd)), sep = " & ")
            row <- paste(row, paste(better.count, " ($", round(100 * better.count / all.count, 2), "\\%$)", sep = ""), sep = " & ")
            row <- paste(row, nan_is_dash(geom.mean(data.mean.ratio[data.mean.ratio < 1])), sep = " & ")
            row <- paste(row, paste(worse.count, " ($", round(100 * worse.count / all.count, 2), "\\%$)", sep = ""), sep = " & ")
            row <- paste(row, nan_is_dash(geom.mean(data.mean.ratio[data.mean.ratio > 1])), sep = " & ")
            cat(paste(escape.latex(row), "\\\\\n"), file = latex.file, append = T)
        }

        cat("\\hline\n", file = latex.file, append = T)
    }

    cat("\\end{tabular}\n", file = latex.file, append = T)
    cat(paste("\\caption{Full ratios.}\n"), file = latex.file, append = T)
    cat("\\end{table}\n", file = latex.file, append = T)

    cat('\\end{landscape}\n', file = latex.file, append = T)
    cat('\\end{document}\n', file = latex.file, append = T)
}
warnings()
