#! /usr/bin/env python3

import sys
import json
import os
import subprocess
import re
import time

################ Timer Begin
start_time = time.time()
############################

def abort(msg):
    print(msg)
    exit(1)

def process_output(output):
    values = []

    for i in range(len(stats)):
        s = stats[i]

        for line in output.stdout.split('\n'):
            if re.search("::" + s + "::", line):
                values.append(line.split("::")[0])
                break

        if len(values) < i + 1:
            values.append("0")

    with open(output_info, "w") as f:
        print("".join(output.stdout), file = f)
        print("".join(output.stderr), file = f)

    return values

################ Main Execution
if len(sys.argv) < 2:
    abort("Usage: " + sys.argv[0] + " <config-file>")

## Loading Config File
config_file = sys.argv[1]

with open(config_file) as f:
    config = json.load(f)

## Getting Config Data
efd_path = config["efd_path"]
bench_path = config["bench_path"] 
architecture_path = config["architecture_path"]
output_qasm_path = config["output_qasm_path"]
output_info_path = config["output_info_path"]
summarized_data_path = config["summarized_data_path"]
allocators = config["allocators"]
runs = config["runs"]
stats = config["stats"]

if not os.path.isfile(efd_path):
    abort("File does not exist: " + efd_path)
if not os.path.isfile(architecture_path):
    abort("File does not exist: " + architecture_path)

if not os.path.isdir(bench_path):
    abort("Folder does not exist: " + bench_path)

if runs["from"] > runs["to"]:
    abort("Starting run must be always less than (or equal) the last one: " +
            "from (" + str(runs["from"]) + ") <= " +
            "to (" + str(runs["to"]) + ")")

summarized_data_file = open(summarized_data_path, "w")

data_header = [
        "Allocator",
        "Benchmark",
        "Run",
        "MaxPartial",
        "MaxChildren",
        "MaxMapseq"
        ] + stats

print(",".join(data_header), file = summarized_data_file)
summarized_data_file.flush()

## Listing Benchmarks
bench_filenames = os.listdir(bench_path)

if len(bench_filenames) == 0:
    abort("No benchmarks found at: " + bench_path)

## Executing Experiment Commands
total_executions = (runs["to"] - runs["from"] + 1) * len(bench_filenames) * len(allocators)
execution_count = 0
success_count = 0

for i in range(runs["from"], runs["to"] + 1):
    i_qasm_path = os.path.join(output_qasm_path, format(i, "02"))
    i_info_path = os.path.join(output_info_path, format(i, "02"))
    os.makedirs(i_qasm_path, exist_ok = True)
    os.makedirs(i_info_path, exist_ok = True)

    for b in bench_filenames:
        b_path = os.path.join(bench_path, b)

        for a in allocators:
            output_qasm = os.path.join(i_qasm_path, b + "." + a["name"] + ".qasm")
            output_info = os.path.join(i_info_path, b + "." + a["name"] + ".info")

            values = [a["name"], b, str(i)]

            cmd = [efd_path,
                    "-i", b_path,
                    "-alloc", a["cmd"],
                    "-arch-file", architecture_path,
                    "-o", output_qasm,
                    "-f", "-stats"]

            for (attr, opt) in [("mp", "partial"), ("mc", "children"), ("ms", "mapseq")]:
                if attr in a:
                    attr_value = str(a[attr])
                    cmd = cmd + ["--bmt-max-" + opt, attr_value]
                    values.append(attr_value)
                else:
                    values.append("0")

            output = subprocess.run(cmd,
                                    stdout = subprocess.PIPE,
                                    stderr = subprocess.PIPE,
                                    universal_newlines = True)

            execution_count += 1
            execution_basic_info = [i, b, a["name"], os.path.basename(architecture_path)]
            percent_info = "(" + str(execution_count) + "/" + str(total_executions) + ")"

            if output.returncode != 0:
                print(percent_info, "[[FAIL]]:", execution_basic_info)
            else:
                print(percent_info, "[[SUCCESS]]:", execution_basic_info)
                success_count += 1

            values = values + process_output(output)
            print(",".join(values), file = summarized_data_file)
            summarized_data_file.flush()

summarized_data_file.close()

################ Timer End
end_time = time.time()
############################

print("Executions Finished:", total_executions)
print("Executions Successfully Completed:", success_count)
print("Total Time:", end_time - start_time, "seconds.")
